import math

import pymunk

import robot, physics

DIE, ATTACK, PATROL, CHASE, IDLE, FINISH = xrange(6)
STATES = 'DIE', 'ATTACK', 'PATROL', 'CHASE', 'IDLE', 'FINISH'


class Enemy(robot.Robot):
    speed = .7 * 60
    detection_radius_sqrd = 6 ** 2
    sight_radius_sqrd = 50 ** 2
    min_dot = math.cos(65 / 57.3)

    sight_filter = pymunk.ShapeFilter(
        mask=pymunk.ShapeFilter.ALL_MASKS ^ (physics.PLAYER + physics.ENEMY + physics.BULLET))
    # sight_filter = pymunk.ShapeFilter(mask=physics.WALL)

    collision_type = physics.ENEMY

    def __init__(self, app):
        robot.Robot.__init__(self, app, (1, 0, 0))

        self.last_node = 0
        self.node = 0
        self.target = None

    def spawn(self, node):
        self.last_node = node
        self.node = node
        self.target = None
        self.health = self.max_health
        self.set_state(PATROL)
        self.body.position = self.app.pathfinder.nodes[self.node][1]
        self.position.xyz = self.body.position.x, 0, self.body.position.y
        if self.body not in self.app.physics.space.bodies:
            self.app.physics.space.add(self.body, self.collider)

    def cache(self):
        if self.body in self.app.physics.space.bodies:
            self.app.physics.space.remove(self.body, self.body.shapes)
        self.app.scene.remove_widget(self)
        self.app.resource_manager.enemy_cache.add(self)

    def set_state(self, state):
        # print 'set state :{}'.format(STATES[state])
        if state == ATTACK:
            self.node = -1
            self.state_timer = self.fire_rate * 2
            p_pos = self.app.game_manager.player.position
            self.app.pathfinder.set_search_area(p_pos.x, p_pos.z)
            self.app.game_manager.refresh_enemy_paths()
            self.set_animation(robot.AIM_ANIM)
            self.head_yaw.angle = -90
            self.head_pitch.angle = 0
        elif state == DIE:
            self.app.physics.space.remove(self.body, self.collider)
            self.set_animation(robot.DEATH_ANIM)
        elif state == IDLE:
            self.state_timer = 2
            self.set_animation(robot.IDLE_ANIM)
        elif state == PATROL:
            self.set_animation(robot.WALK_ANIM)
            self.head_yaw.angle = -90
            self.head_pitch.angle = 0
        elif state == CHASE:
            self.state_timer = 10
            self.node = -1
            self.set_animation(robot.WALK_ANIM)
        elif state == FINISH:
            self.set_animation(robot.IDLE_ANIM)
        self.state = state

    def next_node(self):
        if self.node == -1:
            for node, data in self.app.pathfinder.nodes.iteritems():
                if self.raycast_to(*data[1]):
                    self.node = node
                    self.target = data[1]
                    data[2] = 0
                    break
        else:
            self.last_node = self.node
            self.node = self.app.pathfinder.get_connection(self.node)
            self.target = self.app.pathfinder.nodes[self.node][1]

    def move(self, speed, dt):
        if self.target is None:
            return False
        else:
            dx = self.target[0] - self.position.x
            dy = self.target[1] - self.position.z

            mag_sqrd = dx ** 2 + dy ** 2
            if mag_sqrd < 1:
                self.target = None
                return False
            else:
                mag = mag_sqrd ** .5
                fx = dx / mag * speed * dt
                fy = dy / mag * speed * dt

                self.body.apply_impulse_at_local_point((fx, fy))
                return True

    def can_see(self, x, z):
        dx = x - self.position.x
        dz = z - self.position.z
        mag_sqrd = dx ** 2 + dz ** 2

        if mag_sqrd < self.detection_radius_sqrd:  # too close
            if self.raycast_to(x, z):
                self.target_angle = math.atan2(dz, dx) * -57.3
                return True
        elif mag_sqrd < self.sight_radius_sqrd:  # withing radius
            rads = self.rotation.angle / 57.3
            mag = mag_sqrd ** .5
            dot = ((dx / mag) * math.cos(-rads)) + ((dz / mag) * math.sin(-rads))
            if self.min_dot < dot:  # within fov
                if self.raycast_to(x, z):
                    self.target_angle = math.atan2(dz, dx) * -57.3
                    return True
        return False

    def raycast_to(self, x, z):
        hit = self.app.physics.space.segment_query_first(self.body.position, (x, z), 0, self.sight_filter)
        return hit is None

    def on_damage(self):
        if self.state != FINISH:
            if self.raycast_to(*self.app.game_manager.player.position.xyz[::2]):
                self.set_state(ATTACK)
            else:
                self.set_state(CHASE)

    def update(self, dt, factor=1):
        self.do_animation(dt, factor)
        if self.state == DIE:
            robot.Robot.update(self, dt, factor)
        elif self.state != FINISH:
            can_see = self.can_see(*self.app.game_manager.player.position.xyz[::2])

            if self.state == ATTACK:
                self.target = self.app.game_manager.player.position.xyz[::2]
                if can_see:
                    self.move(self.speed * .5, dt * factor)
                    self.update_shooting(dt * factor)
                else:
                    self.set_state(CHASE)

            elif can_see:
                self.set_state(ATTACK)

            elif self.state == CHASE:
                self.state_timer -= dt
                if self.move(self.speed * 1.5, dt * factor):
                    self.face_velocity(dt)
                elif self.state_timer <= 0:
                    self.set_state(PATROL)
                else:
                    self.next_node()

            elif self.state == PATROL:
                if self.target is None:
                    self.next_node()
                elif self.move(self.speed, dt * factor):
                    self.face_velocity(dt)
                else:
                    self.set_state(IDLE)

            elif self.state == IDLE:
                self.state_timer -= dt
                if self.state_timer <= 0:
                    self.set_state(PATROL)

            robot.Robot.update(self, dt)
